#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Objective', 'simple_objective_rule', '_ObjectiveData', 'minimize', 'maximize', 'simple_objectivelist_rule', 'ObjectiveList']

import sys
import logging
from weakref import ref as weakref_ref
import inspect

from six import iteritems

import pyutilib.math
import pyutilib.misc

from coopr.pyomo.base.numvalue import NumericValue, as_numeric, value, native_numeric_types
from coopr.pyomo.base.expr import _ExpressionBase
from coopr.pyomo.base.component import ActiveComponentData, register_component
from coopr.pyomo.base.sparse_indexed_component import ActiveSparseIndexedComponent, UnindexedComponent_set
from coopr.pyomo.base.misc import apply_indexed_rule, tabular_writer
from coopr.pyomo.base.util import is_functor
from coopr.pyomo.base.sets import Set
from coopr.pyomo.base.var import _VarData
from coopr.pyomo.base.set_types import Reals

logger = logging.getLogger('coopr.pyomo')

_simple_objective_rule_types = set([ type(None), bool ])


# Constants used to define the optimization sense
minimize=1
maximize=-1


def simple_objective_rule( fn ):
    """
    This is a decorator that translates None into Objective.Skip.
    This supports a simpler syntax in objective rules, though these
    can be more difficult to debug when errors occur.

    Example use:

    @simple_objective_rule
    def O_rule(model, i, j):
        ...

    model.o = Objective(rule=simple_objective_rule(...))
    """

    def wrapper_function ( *args, **kwargs ):
        #
        # If the function is None, then skip this objective.
        #
        if fn is None:
            return Objective.Skip
        #
        # Otherwise, the argument is a functor, so call it to generate the 
        # objective expression.
        #
        value = fn( *args, **kwargs )
        if value is None:
            return Objective.Skip
        return value
    return wrapper_function

def simple_objectivelist_rule( fn ):
    """

    This is a decorator that translates None into ObjectiveList.End.
    This supports a simpler syntax in objective rules, though these
    can be more difficult to debug when errors occur.

    Example use:

    @simple_objectivelist_rule
    def O_rule(model, i, j):
        ...

    model.o = ObjectiveList(expr=simple_objectivelist_rule(...))
    """
    def wrapper_function ( *args, **kwargs ):
        #
        # If the function is None, then the list is finished.
        #
        if fn is None:
            return ObjectiveList.End
        #
        # Otherwise, the argument is a functor, so call it to generate the 
        # objective expression.
        #
        value = fn( *args, **kwargs )
        if value is None:
            return ObjectiveList.End
        return value
    return wrapper_function


class _ObjectiveData(ActiveComponentData, NumericValue):
    """
    This class defines the data for a single objective.

    Note that this is a subclass of NumericValue to allow
    objectives to be used as part of expressions.

    Constructor arguments:
        component       The Objective object that owns this data.
        expr            The expression for this objective.

    Public class attributes:
        active          A boolean that is true if this objective is active 
                            in the model.
        expr            The Pyomo expression for this objective
        value           The numeric value of the expression

    Private class attributes:
        _component      The objective component.
    """

    __pickle_slots__ = ( 'value', 'expr')
    __slots__ = __pickle_slots__ + ( '__weakref__', )

    def __init__(self, component, expr):
        # the following lines represent in-lining of the ActiveComponentData
        # and ComponentData constructors.
        self._component = weakref_ref(component)
        self._active = True
        #
        self.value = None
        self.expr = expr

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        state = super(_ObjectiveData, self).__getstate__()
        for i in _ObjectiveData.__pickle_slots__:
            state[i] = getattr(self,i)
        return state

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def __call__(self, exception=True):
        """
        Compute the value of this objective.

	    This method does not simply return self.value because that
	    data value may be out of date w.r.t. the value of decision
	    variables.
        """
        if self.expr is None:
            return None
        return self.expr(exception=exception)

    def polynomial_degree(self):
        """
        Return the polynomial degree of the objective expression.
        """
        if self.expr is None:
            return None
        return self.expr.polynomial_degree()

    def is_minimizing ( self ):
        """
        Return True if this is a minimization objective
        """
        return self.component().sense == minimize


class Objective(ActiveSparseIndexedComponent):
    """
    This modeling component defines an objective expression.

    Note that this is a subclass of NumericValue to allow
    objectives to be used as part of expressions.

    Constructor arguments:
        expr            A Pyomo expression for this objective
        noruleinit      Indicate that its OK that no initialization is
                            specified
        rule            A function that is used to construct objective
                            expressions
        sense           Indicate whether minimizing (the default) or maximizing
        doc             A text string describing this component
        name            A name for this component

    Public class attributes:
        doc             A text string describing this component
        name            A name for this component
        active          A boolean that is true if this component will be 
                            used to construct a model instance
        rule            The rule used to initialize the objective(s)
        sense           The objective sense
        trivial         This boolean is True if all objective indices have 
                            trivial expressions

    Private class attributes:
        _constructed        A boolean that is true if this component has been
                                constructed
        _data               A dictionary from the index set to component data 
                                objects
        _index              The set of valid indices
        _implicit_subsets   A tuple of set objects that represents the index set
        _model              A weakref to the model that owns this component
        _no_rule_init       A boolean that indicates if an initialization rule 
                                is needed
        _parent             A weakref to the parent block that owns this component
        _type               The class type for the derived subclass
    """

    NoObjective = (1000,)
    Skip        = (1000,)

    def __new__(cls, *args, **kwds):
        if cls != Objective:
            return super(Objective, cls).__new__(cls)
        if args == ():
            return SimpleObjective.__new__(SimpleObjective)
        else:
            return IndexedObjective.__new__(IndexedObjective)

    def __init__(self, *args, **kwargs):
        self.sense = kwargs.pop('sense', minimize )
        self.rule  = kwargs.pop('rule', None )
        self._expr  = kwargs.pop('expr', None )
        self._no_rule_init = kwargs.pop('noruleinit', False )
        #
        # This value is set when creating a canonical representation.
        #
        self.trivial = False
        #
        kwargs.setdefault('ctype', Objective)
        ActiveSparseIndexedComponent.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """
        Construct the expression(s) for this objective.
        """
        if __debug__:
            logger.debug("Constructing objective %s", self.cname(True))
        if self._constructed:
            return
        _self_rule = self.rule
        if self._no_rule_init and (_self_rule is not None):
            logger.warning("The noruleinit keyword is being used in conjunction " \
                  "with the rule keyword for objective '%s'; defaulting to " \
                  "rule-based construction" % self.cname(True))
        if _self_rule is None and self._expr is None:
            if not self._no_rule_init:
                logger.warn("No construction rule or expression specified for "
                            "objective '%s'", self.cname(True))
            else:
                self._constructed=True
            return
        self._constructed=True
        #
        _self_parent = self._parent()
        if not self.is_indexed():
            #
            # Scalar component
            #
            if _self_rule is None:
                self.add(None, self._expr)
            else:
                try:
                    tmp = _self_rule(_self_parent)
                except Exception:
                    err = sys.exc_info()[1]
                    logger.error(
                        "Rule failed when generating expression for "
                        "objective %s:\n%s: %s"
                        % ( self.cname(True), type(err).__name__, err ) )
                    raise
                if tmp is None:
                    raise ValueError("Objective rule returned None instead of Objective.Skip")
                self.add(None, tmp)
        else:
            if not self._expr is None:
                raise IndexError("Cannot initialize multiple indices of a constraint with a single expression")
            for val in self._index:
                try:
                    tmp = apply_indexed_rule(self, _self_rule, _self_parent, val)
                except Exception:
                    err = sys.exc_info()[1]
                    logger.error(
                        "Rule failed when generating expression for "
                        "constraint %s with index %s:\n%s: %s"
                        % ( self.cname(True), str(val), type(err).__name__, err ) )
                    raise
                if tmp is None:
                    raise ValueError("Objective rule returned None instead of Objective.Skip for index %s" % str(val))
                self.add(val, tmp)

    def add(self, index, expr):
        """
        Add an objective with a specified index.
        """
        _expr_type = expr.__class__
        if _expr_type is tuple:
            #
            # Ignore an 'empty' objective
            #
            if expr == Objective.Skip:
                return
        elif isinstance(expr, NumericValue):
            #
            # NumericValue -> single Var, single Param, Expression
            #
            pass
        elif _expr_type in native_numeric_types:
            #
            # It is very uncommon case where the objective is simply a number
            # Number -> int, long, float
            #
            expr = as_numeric(expr)
        else:
            raise ValueError("Bad objective expression: %s" % str(expr))
        #
        # Add the objective data
        #
        if index is None:
            self._data[index] = self
            self.expr = expr
        else:
            self._data[index] = _ObjectiveData(self, expr)

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        return ( 
            [("Size", len(self)),
             ("Index", self._index \
                       if self._index != UnindexedComponent_set else None),
             ("Active", self.active),
             ("Sense", "minimize" if self.sense == minimize else "maximize"),
             ],
            iteritems(self._data),
            ( "Key","Active","Expression" ),
            lambda k, v: [ k,
                           v.active,
                           v.expr
                           ]
            )

    def display(self, prefix="", ostream=None): 
        """Provide a verbose display of this object""" 
        if not self.active: 
            return 
        tab = "    " 
        if ostream is None: 
            ostream = sys.stdout 
        ostream.write(prefix+self.cname()+" : ") 
        ostream.write(", ".join("%s=%s" % (k,v) for k,v in [
                    ("Size", len(self)),
                    ("Index", self._index \
                         if self._index != UnindexedComponent_set else None),
                    ("Active", self.active),
                    ] ))
        
        ostream.write("\n") 
        tabular_writer( ostream, prefix+tab,  
                        ((k,v) for k,v in iteritems(self._data) if v.active), 
                        ( "Key","Active","Value" ), 
                        lambda k, v: [ k, v.active, value(v), ] ) 


class SimpleObjective(Objective, _ObjectiveData):
    """
    SimpleObjective is the implementation representing a single,
    non-indexed objective.
    """

    def __init__(self, *args, **kwd):
        _ObjectiveData.__init__(self, self, None)
        Objective.__init__(self, *args, **kwd)
        self._data = {}

    def __call__(self, exception=True):
        """
        Compute the value of the objective expression
        """
        if self._constructed:
            return _ObjectiveData.__call__(self, exception=exception)
        if exception:
            raise ValueError( """Evaluating the numeric value of objective '%s' before the Objective has been
            constructed (there is currently no value to return).""" % self.cname(True) )


    #
    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, we do not
    # need to define the __getstate__ or __setstate__ methods.
    # We just defer to the super() get/set state.  Since all of our 
    # get/set state methods rely on super() to traverse the MRO, this 
    # will automatically pick up both the Component and Data base classes.
    #


class IndexedObjective(Objective):

    def __call__(self, exception=True):
        """Compute the value of the objective body"""
        if exception:
            msg = 'Cannot compute the value of an array of objectives'
            raise TypeError(msg)

    def is_minimizing ( self ):
        """Return true if this is a minimization objective"""
        return self.sense == minimize


class ObjectiveList(IndexedObjective):
    """
    An objective component that represents a list of objectives.
    Objectives can be indexed by their index, but when they are added
    an index value is not specified.
    """

    End             = (1003,)

    def __init__(self, **kwargs):
        """Constructor"""
        args = (Set(),)
        self._nobjectives = 0
        Objective.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """
        Construct the expression(s) for this objective.
        """
        if __debug__:
            logger.debug("Constructing objective %s", self.cname(True))
        if self._constructed:
            return
        _self_rule = self.rule
        if self._no_rule_init and (_self_rule is not None):
            logger.warning("The noruleinit keyword is being used in conjunction " \
                  "with the rule keyword for objective '%s'; defaulting to " \
                  "rule-based construction" % self.cname(True))
        self._constructed=True
        if _self_rule is None:
            if not self._no_rule_init:
                logger.warn("No construction rule or expression specified for "
                            "objective '%s'", self.cname(True))
            return
        #
        _generator = None
        _self_parent = self._parent()
        if inspect.isgeneratorfunction(_self_rule):
            _generator = _self_rule(_self_parent)
        elif inspect.isgenerator(_self_rule):
            _generator = _self_rule
        if _generator is None:
            while True:
                val = self._nobjectives + 1
                if __debug__ is True:
                    logger.debug("   Constructing objective index "+str(val))
                expr = apply_indexed_rule( self, _self_rule, _self_parent, val )
                if expr is None:
                    raise ValueError( "Objective rule returned None "
                                      "instead of ObjectiveList.End" )
                if (expr.__class__ is tuple and expr == ObjectiveList.End):
                    return
                self.add(expr)
        else:
            for expr in _generator:
                if expr is None:
                    raise ValueError( "Objective generator returned None "
                                      "instead of ObjectiveList.End" )
                if (expr.__class__ is tuple and expr == ObjectiveList.End):
                    return
                self.add(expr)

    def add(self, expr):
        """
        Add an objective to the list.
        """
        self._nobjectives += 1
        Objective.add(self, self._nobjectives, expr)


register_component(Objective, 'Expressions that are minimized or maximized in a model.')
register_component(ObjectiveList, "A list of objectives in a model.")



#
# Unit Tests for BuildCheck() Objects
#
# PyomoModel                Base test class
# Simple                    Test scalar parameter
# Array1                    Test arrays of parameters
#

import unittest
import os
import sys
from os.path import abspath, dirname
from coopr.pyomo import *

class PyomoModel(unittest.TestCase):

    def setUp(self):
        self.model = AbstractModel()

    def construct(self,filename):
        self.instance = self.model.create(filename)


def action1a_fn(model):
    return value(model.A) == 3.3

def action1b_fn(model):
    return value(model.A) != 3.3

def action2a_fn(model, i):
    ans=True
    if i in model.A:
        return (value(model.A[i]) == 1.3)
    return True

def action2b_fn(model, i):
    if i in model.A:
        ans = (value(model.A[i]) == 1.3)
        #print "HERE",i,ans,not ans
        return not ans
    return True


class Simple(PyomoModel):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.A = Param(initialize=3.3)

    def tearDown(self):
        if os.path.exists("param.dat"):
            os.remove("param.dat")

    def test_true(self):
        """Apply a build check that returns true"""
        self.model.action1 = BuildCheck(rule=action1a_fn)
        self.instance = self.model.create()
        tmp = value(self.instance.A)
        self.assertEqual( tmp, 3.3 )

    def test_false(self):
        """Apply a build check that returns false"""
        self.model.action1 = BuildCheck(rule=action1b_fn)
        try:
            self.instance = self.model.create()
            self.fail("expected failure")
        except ValueError:
            pass



class Array1(PyomoModel):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.Z = Set(initialize=[1,3])
        self.model.A = Param(self.model.Z, initialize=1.3)

    def test_true(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2a_fn)
        self.instance = self.model.create()

    def test_false(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2b_fn)
        try:
            self.instance = self.model.create()
            self.fail("expected failure")
        except ValueError:
            pass


class Array2(PyomoModel):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.Z = Set(initialize=[1,3])
        self.model.A = Param(self.model.Z, initialize=1.3)

    def test_true(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2a_fn)
        self.instance = self.model.create()

    def test_false(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2b_fn)
        try:
            self.instance = self.model.create()
            self.fail("expected failure")
        except ValueError:
            pass


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for Elements of a Model
#
# TestSimpleCon                Class for testing single constraint
# TestArrayCon                Class for testing array of constraint
#

import logging
import os
import sys

from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo import *
from coopr.opt import *
from coopr.pyomo.base.connector import ConnectorExpander
import pyutilib.th as unittest
import pyutilib.services


class TestBlockModel(unittest.TestCase):
    def test_nonindexed_block_immutable_param(self):
        model = AbstractModel()
        def _b_rule(b):
            b.A = Param(initialize=2.0)
        model.B = Block(rule=_b_rule)

        instance = model.create()

        self.assertEqual(value(instance.B.A), 2.0)

    def test_nonindexed_block_mutable_param(self):
        model = AbstractModel()
        def _b_rule(b):
            b.A = Param(initialize=2.0, mutable=True)
        model.B = Block(rule=_b_rule)

        instance = model.create()
        self.assertEqual(value(instance.B.A), 2.0)

        instance.B.A = 4.0
        instance.preprocess()
        self.assertEqual(value(instance.B.A), 4.0)


    def test_indexed_block_immutable_param(self):
        model = AbstractModel()
        model.A = RangeSet(2)
        def _b_rule(b,id):
            b.A = Param(initialize=id)
        model.B = Block(model.A, rule=_b_rule)

        instance = model.create()

        self.assertEqual(value(instance.B[1].A),1)
        self.assertEqual(value(instance.B[2].A),2)

    def test_indexed_block_mutable_param(self):
        model = AbstractModel()
        model.A = RangeSet(2)
        def _b_rule(b,id):
            b.A = Param(initialize=id, mutable=True)
        model.B = Block(model.A, rule=_b_rule)

        instance = model.create()

        self.assertEqual(value(instance.B[1].A),1)
        self.assertEqual(value(instance.B[2].A),2)

        instance.B[1].A = 4.0
        instance.preprocess()
        self.assertEqual(value(instance.B[1].A),4.0)


    def test_create_from_dict(self):
        model = AbstractModel()
        model.A = RangeSet(2)
        def _b_rule(b,id):
            b.S = Set()
            b.P = Param()
            b.Q = Param(b.S)
        model.B = Block(model.A, rule=_b_rule)

        instance = model.create( {None:{'B': \
                                           {1:{'S':{None:['a','b','c']}, \
                                               'P':{None:4}, \
                                               'Q':{('a',):1,('b',):2,('c',):3}}, \
                                            2:{'S':{None:[]}, \
                                               'P':{None:3}} \
                                           } \
                                        }} ) 

        self.assertEqual(set(instance.B[1].S),set(['a','b','c']))
        self.assertEqual(value(instance.B[1].P),4)
        self.assertEqual(value(instance.B[1].Q['a']),1)
        self.assertEqual(value(instance.B[1].Q['b']),2)
        self.assertEqual(value(instance.B[1].Q['c']),3)
        self.assertEqual(value(instance.B[2].P),3)

    def test_expand_connector(self):
        model = AbstractModel()
        model.A = Set()
        def _b_rule(b,id):
            b.X = Var()
            b.PORT = Connector()
            b.PORT.add( b.X )
        model.B = Block(model.A, rule=_b_rule)

        def _c_rule(m,a):
            return m.B[a].PORT == m.B[(a+1)%2].PORT
        model.C = Constraint(model.A,rule=_c_rule)

        instance = model.create( {None: {'A':{None:[0,1]}}} )

        # FIXME: Not sure what to assert here, but at the moment this throws an error anyways. 


if __name__ == "__main__":
    unittest.main()




#
# Unit Tests for BuildAction() Objects
#
# PyomoModel                Base test class
# Simple                    Test scalar parameter
# Array1                    Test arrays of parameters
#

import unittest
import os
import sys
from os.path import abspath, dirname
from coopr.pyomo import *


def action1_fn(model):
    model.A = 4.3

def action2_fn(model, i):
    if i in model.A:
        model.A[i] = value(model.A[i])+i

def action3_fn(model, i):
    if i in model.A.sparse_keys():
        model.A[i] = value(model.A[i])+i


class Simple(unittest.TestCase):

    def setUp(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.A = Param(initialize=3.3, mutable=True)
        model.action1 = BuildAction(rule=action1_fn)
        self.instance = model.create()

    def tearDown(self):
        self.instance = None
        if os.path.exists("param.dat"):
            os.remove("param.dat")

    def test_value(self):
        """Check the value of the parameter"""
        tmp = value(self.instance.A.value)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 4.3 )
        self.assertEqual(value(self.instance.A.value), value(self.instance.A))

    def test_getattr(self):
        """Check the use of the __getattr__ method"""
        self.assertEqual( self.instance.A.value, 4.3)



class Array_Param(unittest.TestCase):

    def test_sparse_param_nodefault(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1,3])
        model.A = Param(model.Z, initialize={1:1.3}, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create()

        tmp = value(instance.A[1])
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 2.3 )

    def test_sparse_param_nodefault_sparse_iter(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1,3])
        model.A = Param(model.Z, initialize={1:1.3}, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action3_fn)
        instance = model.create()

        tmp = value(instance.A[1])
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 2.3 )

    def test_sparse_param_default(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1,3])
        model.A = Param(model.Z, initialize={1:1.3}, default=0, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create()

        tmp = value(instance.A[1])
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 2.3 )


    def test_dense_param(self):
        #
        # Create model instance
        #
        model = AbstractModel()
        model.Z = Set(initialize=[1,3])
        model.A = Param(model.Z, initialize=1.3, mutable=True)
        model.action2 = BuildAction(model.Z, rule=action2_fn)
        instance = model.create()

        self.assertEqual( instance.A[1], 2.3)
        self.assertEqual( value(instance.A[3]), 4.3)


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

import sys
import logging

import coopr.pyomo
from coopr.pyomo.base import Constraint, \
                             Objective, \
                             Suffix, \
                             active_subcomponents_data_generator
from coopr.pyomo.expr import generate_ampl_repn


def preprocess_block_objectives(block):

    # Create a suffix component for the repn
    block_repn_suffix = block.subcomponent("_ampl_repn")
    if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
        block.del_component("_ampl_repn")
        block_repn_suffix = block._ampl_repn = Suffix(direction=Suffix.LOCAL,datatype=None)

    all_repns = list()
    for objective_data in active_subcomponents_data_generator(block, Objective): #recursive = False

        if objective_data.expr is None:
            raise ValueError("No expression has been defined for objective %s" % str(key))

        try:
            ampl_repn = generate_ampl_repn(objective_data.expr)
        except Exception:
            err = sys.exc_info()[1]
            logging.getLogger('coopr.pyomo').error\
                ( "exception generating a ampl representation for objective %s: %s" \
                      % (objective_data.cname(True), str(err)) )
            raise
            
        all_repns.append((objective_data,ampl_repn))

    if len(all_repns) > 0:
        # expand=False because we know there are no Array components in this list
        # and this provides a performance improvement
        block_repn_suffix.updateValues(all_repns, expand=False)

def preprocess_block_constraints(block):

    # Create a suffix component for the repn
    block_repn_suffix = block.subcomponent("_ampl_repn")
    if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
        block.del_component("_ampl_repn")
        block_repn_suffix = block._ampl_repn = Suffix(direction=Suffix.LOCAL,datatype=None)

    all_repns = list()
    for constraint_data in active_subcomponents_data_generator(block, Constraint): #recursive = False

        if constraint_data.body is None:
            raise ValueError("No expression has been defined for the body of constraint %s, index=%s" % (str(constraint.name), str(index)))

        try:
            ampl_repn = generate_ampl_repn(constraint_data.body)
        except Exception:
            err = sys.exc_info()[1]
            logging.getLogger('coopr.pyomo').error\
                ( "exception generating a ampl representation for constraint %s: %s" \
                      % (constraint_data.cname(True), str(err)) )
            raise
        
        all_repns.append((constraint_data,ampl_repn))
                
    if len(all_repns) > 0:
        # expand=False because we know there are no Array components in this list
        # and this provides a performance improvement
        block_repn_suffix.updateValues(all_repns, expand=False)

@coopr.core.coopr_api(namespace='pyomo.model')
def compute_ampl_repn(data, model=None):
    """
    This plugin computes the ampl representation for all
    objectives and constraints. All results are stored at the
    block level on a Suffix named "_ampl_repn"

    NOTE: this does not check for trivial constraints

    We break out preprocessing of the objectives and constraints
    in order to avoid redundant and unnecessary work, specifically
    in contexts where a model is iteratively solved and modified.
    we don't have finer-grained resolution, but we could easily
    pass in a Constraint and an Objective if warranted.

    Required:
        model:      A concrete model instance.
    """
    for block in model.all_blocks():
        preprocess_block_constraints(block)
        preprocess_block_objectives(block)




import copy

import pyutilib.th as unittest
from coopr.pyomo import *
#from coopr.pyomo.base.expression import _ExpressionData
from six import StringIO

class TestExpressionData(unittest.TestCase):
    
    # The copy method must be invoked on expression container to obtain
    # a shollow copy of the class, the underlying expression remains
    # a reference.
    def test_copy(self):
        model = ConcreteModel()
        model.a = Var(initialize=5)
        model.b = Var(initialize=10)

        model.expr1 = Expression(initialize=model.a)
        
        # Do a shallow copy, the same underlying expression is still referenced
        expr2 = copy.copy(model.expr1)
        self.assertEqual( model.expr1(), 5 )
        self.assertEqual( expr2(), 5 )
        self.assertEqual( id(model.expr1.value), id(expr2.value) )

        # Do an in place modification the expression
        model.expr1.value.value = 1
        self.assertEqual( model.expr1(), 1 )
        self.assertEqual( expr2(), 1 )
        self.assertEqual( id(model.expr1.value), id(expr2.value) )

        # Update the expression value on expr1 only
        model.expr1.value = model.b
        self.assertEqual( model.expr1(), 10 )
        self.assertEqual( expr2(), 1 )
        self.assertNotEqual( id(model.expr1.value), id(expr2.value) )

        model.a.value = 5
        model.b.value = 10
        model.del_component('expr1')
        model.expr1 = Expression(initialize=model.a + model.b)

        # Do a shallow copy, the same underlying expression is still referenced
        expr2 = copy.copy(model.expr1)
        self.assertEqual( model.expr1(), 15 )
        self.assertEqual( expr2(), 15 )
        self.assertEqual( id(model.expr1.value), id(expr2.value) )
        self.assertEqual( id(model.expr1.value._args[0]),
                          id(expr2.value._args[0]) )
        self.assertEqual( id(model.expr1.value._args[1]),
                          id(expr2.value._args[1]) )


        # Do an in place modification the expression
        # This causes cloning due to reference counting
        model.a.value = 0
        self.assertEqual( model.expr1(), 10 )
        self.assertEqual( expr2(), 10 )
        self.assertEqual( id(model.expr1.value), id(expr2.value) )
        self.assertEqual( id(model.expr1.value._args[0]),
                          id(expr2.value._args[0]) )
        self.assertEqual( id(model.expr1.value._args[1]),
                          id(expr2.value._args[1]) )


        # Do an in place modification the expression
        # This causes cloning due to reference counting
        model.expr1.value += 1
        self.assertEqual( model.expr1(), 11 )
        self.assertEqual( expr2(), 10 )
        self.assertNotEqual( id(model.expr1.value), id(expr2.value) )

    # test that an object is properly deepcopied when the model is cloned
    def test_model_clone(self):
        model = ConcreteModel()
        model.x = Var(initialize=2.0)
        model.y = Var(initialize=0.0)
        model.ec = Expression(initialize=model.x**2+1)
        model.obj = Objective(expr=model.y+model.ec)
        self.assertEqual(model.obj.expr(),5.0)
        self.assertTrue(id(model.ec) in [id(e) for e in model.obj.expr._args])
        inst = model.clone()
        self.assertEqual(inst.obj.expr(),5.0)
        self.assertTrue(id(inst.ec) in [id(e) for e in inst.obj.expr._args])
        self.assertNotEqual(id(model.ec), id(inst.ec))
        self.assertFalse(id(inst.ec) in [id(e) for e in model.obj.expr._args])

    def test_is_constant(self):
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.p = Param(initialize=1.0)
        model.ec = Expression(initialize=model.x)
        self.assertEqual(model.ec.is_constant(), False)
        self.assertEqual(model.ec.value.is_constant(), False)
        model.ec.value = model.p
        self.assertEqual(model.ec.is_constant(), False)
        self.assertEqual(model.ec.value.is_constant(), True)

    def test_polynomial_degree(self):
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=model.x)
        self.assertEqual( model.ec.polynomial_degree(), 
                          model.ec.value.polynomial_degree() )
        self.assertEqual(model.ec.polynomial_degree(), 1)
        model.ec.value = model.x**2
        self.assertEqual( model.ec.polynomial_degree(), 
                          model.ec.value.polynomial_degree())
        self.assertEqual( model.ec.polynomial_degree(), 2 )


    def test_init_concrete(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)

        model.ec = Expression(initialize=0)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=model.y)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.y.value = -1
        model.ec = Expression(initialize=model.y+1.0)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

    def test_init_abstract(self):
        model = AbstractModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

        model.del_component('obj')
        model.ec = Expression(initialize=0.0)
        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

        model.del_component('obj')
        model.ec = Expression(initialize=0.0)
        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

class TestExpression(unittest.TestCase):

    def setUp(self):
        TestExpression._save = coopr.pyomo.base.expr.TO_STRING_VERBOSE
        # Tests can choose what they want - this just makes sure that
        #things are restored after the tests run.
        #coopr.pyomo.base.expr.TO_STRING_VERBOSE = True

    def tearDown(self):
        coopr.pyomo.base.expr.TO_STRING_VERBOSE = TestExpression._save

    def test_init_concrete_indexed(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var([1,2,3],initialize=1.0)

        model.ec = Expression([1,2,3],initialize=1.0)
        model.obj = Objective(expr=1.0+summation(model.ec, index=[1,2,3]))
        self.assertEqual(model.obj.expr(),4.0)
        model.ec[1].value = 2.0
        self.assertEqual(model.obj.expr(),5.0)

    def test_init_concrete_nonindexed(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)

        model.ec = Expression(initialize=0)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=model.y)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.y.value = -1
        model.ec = Expression(initialize=model.y+1.0)
        model.obj = Objective(expr=1.0+model.ec)
        self.assertEqual(model.obj.expr(),1.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e = 1.0
        model.ec.value = e
        self.assertEqual(model.obj.expr(),2.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        model.ec.value = e
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(),3.0)
        self.assertEqual(id(model.obj.expr._args[0]),id(model.ec))

    def test_init_abstract_indexed(self):
        model = AbstractModel()
        model.ec = Expression([1,2,3],initialize=1.0)
        model.obj = Objective(rule=lambda m: 1.0+summation(m.ec,index=[1,2,3]))
        inst = model.create()
        self.assertEqual(inst.obj.expr(),4.0)
        inst.ec[1].value = 2.0
        self.assertEqual(inst.obj.expr(),5.0)

    def test_init_abstract_nonindexed(self):
        model = AbstractModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)
        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)
        def obj_rule(model):
            return 1.0+model.ec
        model.obj = Objective(expr=obj_rule)
        inst = model.create()
        self.assertEqual(inst.obj.expr(),1.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e = 1.0
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),2.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        inst.ec.value = e
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(),3.0)
        self.assertEqual(id(inst.obj.expr._args[0]),id(inst.ec))

    def test_pprint_oldStyle(self):
        coopr.pyomo.base.expr.TO_STRING_VERBOSE = True

        model = ConcreteModel()
        model.x = Var()
        model.e = Expression(initialize=model.x+2.0)
        model.E = Expression([1,2],initialize=model.x**2+1)
        expr = model.e*model.x**2 + model.E[1]

        output = \
"""\
sum( prod( num=( e{sum( 2.0 , x )} , pow( x , 2.0 ) ) ) , E[1]{sum( 1 , pow( x , 2.0 ) )} )
e : Size=1, Index=None, Default=None
    Key  : Expression
    None : sum( 2.0 , x )
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : sum( 1 , pow( x , 2.0 ) )
      2 : sum( 1 , pow( x , 2.0 ) )
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.value = 1.0
        model.E[1].value = 2.0
        output = \
"""\
sum( prod( num=( e{1.0} , pow( x , 2.0 ) ) ) , E[1]{2.0} )
e : Size=1, Index=None, Default=None
    Key  : Expression
    None :        1.0
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : 2.0
      2 : sum( 1 , pow( x , 2.0 ) )
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())


        model.e.value = None
        model.E[1].value = None
        output = \
"""\
sum( prod( num=( e{Undefined} , pow( x , 2.0 ) ) ) , E[1]{Undefined} )
e : Size=1, Index=None, Default=None
    Key  : Expression
    None :  Undefined
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : Undefined
      2 : sum( 1 , pow( x , 2.0 ) )
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())


    def test_pprint_newStyle(self):
        coopr.pyomo.base.expr.TO_STRING_VERBOSE = False

        model = ConcreteModel()
        model.x = Var()
        model.e = Expression(initialize=model.x+2.0)
        model.E = Expression([1,2],initialize=model.x**2+1)
        expr = model.e*model.x**2 + model.E[1]

        output = \
"""\
( 2.0 + x ) * x**2.0 + 1 + x**2.0
e : Size=1, Index=None, Default=None
    Key  : Expression
    None : 2.0 + x
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : 1 + x**2.0
      2 : 1 + x**2.0
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.value = 1.0
        model.E[1].value = 2.0
        output = \
"""\
1.0 * x**2.0 + 2.0
e : Size=1, Index=None, Default=None
    Key  : Expression
    None :        1.0
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : 2.0
      2 : 1 + x**2.0
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())


        model.e.value = None
        model.E[1].value = None
        output = \
"""\
Undefined * x**2.0 + Undefined
e : Size=1, Index=None, Default=None
    Key  : Expression
    None :  Undefined
E : Size=2, Index=E_index, Default=None
    Key : Expression
      1 : Undefined
      2 : 1 + x**2.0
"""
        out = StringIO()
        out.write(str(expr)+"\n")
        model.e.pprint(ostream=out)
        #model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())


if __name__ == "__main__":
    unittest.main()


    



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = [ 'value', 'NumericValue', 'as_numeric', 'NumericConstant',
            'is_constant', 'is_fixed']

import sys
import logging
from six import iteritems, PY3, string_types, text_type, binary_type

from coopr.pyomo.base.set_types import Reals, Any

logger = logging.getLogger('coopr.pyomo')

def create_name(name, ndx):
    """
    Create a canonical name for a component using the given index.
    """
    if ndx is None:
        return name
    if type(ndx) is tuple:
        tmp = str(ndx).replace(', ',',')
        return name+"["+tmp[1:-1]+"]"
    return name+"["+str(ndx)+"]"


##------------------------------------------------------------------------
##
## Standard types of expressions
##
##------------------------------------------------------------------------


def _old_value(obj):
    """
    A utility function that returns the value of a Pyomo object or expression.

    If the argument is None, a numeric value or a string, then this
    function simply returns the argument.  Otherwise, if the argument is
    a NumericValue then the __call__ method is executed.
    """
    if obj is None:
        return None
    if type(obj) in (bool,int,long,float,str):
        return obj
    if not isinstance(obj, NumericValue):
        raise ValueError("Object %s is not a NumericValue object" % (obj,))
    tmp = obj()
    if tmp is None:
        raise ValueError("No value for uninitialized NumericValue object %s"
                         % (obj.cname(),))
    return tmp

# It is *significantly* faster to build the list of types we want to
# test against as a "static" set, and not to regenerate it locally for
# every call.  Plus, this allows us to dynamically augment the set with
# new "native" types (e.g., from NumPy)
native_numeric_types = set([ int, float, bool ])
try:
    native_numeric_types.add(long)
except:
    pass
native_types = set([ bool, str, type(None) ]).union( native_numeric_types )
if PY3:
    native_types.add(bytes)
else:
    native_types.add(unicode)

def value(obj, exception=True):
    """
    A utility function that returns the value of a Pyomo object or expression.

    If the argument is None, a numeric value or a string, then this
    function simply returns the argument.  Otherwise, if the argument is
    a NumericValue then the __call__ method is executed.
    """
    if obj.__class__ in native_types:
        return obj
    try:
        numeric = obj.as_numeric()
    except AttributeError:
        try:
            numeric = as_numeric(obj)
        except ValueError:
            if isinstance( obj, string_types + (text_type, binary_type) ):
                native_types.add(type(obj))
                return obj
            raise
    try:
        tmp = numeric(exception=exception)
    except:
        logger.error(
            "evaluating object as numeric value: %s\n    (object: %s)\n%s"
            % (obj, type(obj), sys.exc_info()[1]))
        raise
    
    if exception and (tmp is None):
        raise ValueError("No value for uninitialized NumericValue object %s"
                         % (obj.cname(),))
    return tmp


def is_constant(obj):
    """
    A utility function that returns a boolean that indicates whether the
    object is a constant
    """
    # This method is rarely, if ever, called.  Plus, since the
    # expression generation (and constraint generation) system converts
    # everything to NumericValues, it is better (i.e., faster) to assume
    # that the obj is a NumericValue
    #
    # JDS: NB: I am not sure why we allow str to be a constant, but
    # since we have historically done so, we check for type membership
    # in native_types and not in native_numeric_types.
    #
    if obj.__class__ in native_types:
        return True
    try:
        return obj.is_constant()
    except AttributeError:
        pass
    return as_numeric(obj).is_constant()

def is_fixed(obj):
    """
    A utility function that returns a boolean that indicates whether the
    input object's value is fixed.
    """
    # JDS: NB: I am not sure why we allow str to be a constant, but
    # since we have historically done so, we check for type membership
    # in native_types and not in native_numeric_types.
    #
    if obj.__class__ in native_types:
        return True
    try:
        return obj.is_fixed()
    except AttributeError:
        pass
    return as_numeric(obj).is_fixed()


# It is very common to have only a few constants in a model, but those
# constants get repeated many times.  KnownConstants lets us re-use /
# share constants we have seen before.
KnownConstants = {}

def as_numeric(obj):
    """
    Verify that this obj is a NumericValue or intrinsic value.
    """
    # int and float are *so* common that it pays to treat them specially
    if obj.__class__ in native_numeric_types:
        if obj in KnownConstants:
            return KnownConstants[obj]
        else:
            # Because INT, FLOAT, and sometimes LONG hash the same, we
            # want to convert them to a common type (at the very least,
            # so that the order in which tests run does not change the
            # results!)
            try:
                tmp = float(obj)
                if tmp == obj:
                    tmp = NumericConstant(tmp)
                    KnownConstants[obj] = tmp
                    return tmp
            except:
                pass

            tmp = NumericConstant(obj)
            KnownConstants[obj] = tmp
            return tmp
    try:
        return obj.as_numeric()
    except AttributeError:
        pass
    try: 
        if obj.__class__ is (obj + 0).__class__:
            # obj may (or may not) be hashable, so we need this try
            # block so that things proceed normally for non-hashable
            # "numeric" types
            try:
                if obj in KnownConstants:
                    return KnownConstants[obj]
                else:
                    tmp = NumericConstant(obj)
                    KnownConstants[obj] = tmp
                    
                    # If we get here, this is a reasonably well-behaving
                    # numeric type: add it to the native numeric types
                    # so that future lookups will be faster.
                    native_numeric_types.add(obj.__class__)
                    # native numeric types are also native types
                    native_types.add(obj.__class__)

                    return tmp
            except:
                return NumericConstant(obj)
    except:
        pass
    raise ValueError(
        "Cannot convert object of type '%s' (value = %s) to a numeric value." 
        % (type(obj).__name__, obj, ))


class NumericValue(object):
    """
    This is the base class for numeric values used in Pyomo.

    For efficiency purposes, some derived classes do not call this
    constructor (e.g. see the "_ExpressionBase" class defined in "expr.py").
    This is fine if the value is finite or unspecified.  In that
    case, no validation is performed, so the subclass can simply
    specify the name, domain and value.

    Constructor Arguments:
        domain          The value domain.  Unspecified default: None.
        name            The value name.  Unspecified default: None
        value           The initial value.  Unspecified default: None.

    Public Class Attributes:
        value           The numeric value of this object.
    """

    __slots__ = ()

    # This is required because we define __eq__
    __hash__ = None

    def __getstate__(self):
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__slots__:
        #    state[i] = getattr(self,i)
        # return state
        #
        # However, in this case, the (nominal) parent class is 'object',
        # and object does not implement __getstate__.  So, we will check
        # to make sure that there is a base __getstate__() to call...
        #
        # Further, since there are actually no slots defined here, the
        # real question is to either return an empty dict or the
        # parent's dict.
        _base = super(NumericValue, self)
        if hasattr(_base, '__getstate__'):
            return _base.__getstate__()
        else:
            return {}

    def __setstate__(self, state):
        # Note: our model for setstate is for derived classes to modify
        # the state dictionary as control passes up the inheritance
        # hierarchy (using super() calls).  All assignment of state ->
        # object attributes is handled at the last class before 'object'
        # (which may -- or may not (thanks to MRO) -- be here.
        _base = super(NumericValue, self)
        if hasattr(_base, '__setstate__'):
            return _base.__setstate__(state)
        else:
            for key, val in iteritems(state):
                # Note: per the Python data model docs, we explicitly
                # set the attribute using object.__setattr__() instead
                # of setting self.__dict__[key] = val.
                object.__setattr__(self, key, val)
            

    def cname(self, fully_qualified=False, name_buffer=None):
        """TODO"""
        _base = super(NumericValue, self)
        if hasattr(_base,'cname'):
            return _base.cname(fully_qualified, name_buffer)
        else:
            return str(type(self))

    def is_constant(self):
        """Return True if this numeric value is a constant value."""
        return False

    def is_fixed(self):
        """Return True is this is a non-constant value that has been fixed."""
        return False

    def is_expression(self):
        """Return True if this numeric value is an expression."""
        return False

    def is_relational(self):
        """
        Return True if this numeric value represents a relational expression.
        """
        return False

    def is_indexed(self):
        """Return True if this numeric value is an indexed object."""
        return False

    def as_numeric(self):
        return self

    def polynomial_degree(self):
        """Return the polynomial degree of this expression."""
        return 0

    def reset(self):            #pragma:nocover
        """Reset the value of this numeric object"""
        pass                    #pragma:nocover

    #def set_value(self, val):
    #    """Set the value of this numeric object, after validating its value."""
    #    if self._valid_value(val):
    #        self.value=val

    #def __nonzero__(self):
    #    """Return True if the value is defined and non-zero."""
    #    if self.value is None:
    #        raise ValueError("Numeric value is undefined")
    #    if self.value:
    #        return True
    #    return False

    #def __call__(self, exception=True):
    #    """Return the value of this object."""
    #    return self.value

    def __float__(self):
        """Coerce the value to a floating point."""
        tmp = self.__call__()
        if tmp is None:
            raise ValueError("Cannot coerce numeric value `%s' to float "
                             "because it is uninitialized." % (self.cname(),))
        return float(tmp)

    def __int__(self):
        """Coerce the value to an integer point."""
        tmp = self.__call__()
        if tmp is None:
            raise ValueError("Cannot coerce numeric value `%s' to integer "
                             "because it is uninitialized." % (self.cname(),))
        return int(tmp)

    #def _valid_value(self, value, use_exception=True):
    #    """
    #    Validate the value.  If use_exception is True, then raise an
    #    exception.
    #    """
    #    ans = value is None or self.domain is None or value in self.domain
    #    if not ans and use_exception:
    #        raise ValueError("Numeric value `%s` is not in domain %s"
    #                         % (value, self.domain))
    #    return ans

    def __lt__(self,other):
        """Less than operator

        (Called in response to 'self < other' or 'other > self'.)
        """
        return generate_relational_expression(_lt, self, other)

    def __gt__(self,other):
        """Greater than operator

        (Called in response to 'self > other' or 'other < self'.)
        """
        return generate_relational_expression(_lt, other, self)

    def __le__(self,other):
        """Less than or equal operator

        (Called in response to 'self <= other' or 'other >= self'.)
        """
        return generate_relational_expression(_le, self, other)

    def __ge__(self,other):
        """Greater than or equal operator

        (Called in response to 'self >= other' or 'other <= self'.)
        """
        return generate_relational_expression(_le, other, self)

    def __eq__(self,other):
        """Equal to operator

        (Called in response to 'self = other'.)
        """
        return generate_relational_expression(_eq, self, other)

    def __add__(self,other):
        """Binary addition

        (Called in response to 'self + other'.)
        """
        return generate_expression(_add,self,other)

    def __sub__(self,other):
        """ Binary subtraction

        (Called in response to 'self - other'.)
        """
        return generate_expression(_sub,self,other)

    def __mul__(self,other):
        """ Binary multiplication

        (Called in response to 'self * other'.)
        """
        return generate_expression(_mul,self,other)

    def __div__(self,other):
        """ Binary division

        (Called in response to 'self / other'.)
        """
        return generate_expression(_div,self,other)

    def __truediv__(self,other):
        """ Binary division

        (Called in response to 'self / other' with __future__.division.)
        """
        return generate_expression(_div,self,other)

    def __pow__(self,other):
        """ Binary power

        (Called in response to 'self ** other'.)
        """
        return generate_expression(_pow,self,other)

    def __radd__(self,other):
        """Binary addition

        (Called in response to 'other + self'.)
        """
        return generate_expression(_radd,self,other)

    def __rsub__(self,other):
        """ Binary subtraction

        (Called in response to 'other - self'.)
        """
        return generate_expression(_rsub,self,other)

    def __rmul__(self,other):
        """ Binary multiplication

        (Called in response to 'other * self'.)
        """
        return generate_expression(_rmul,self,other)

    def __rdiv__(self,other):
        """ Binary division

        (Called in response to 'other / self'.)
        """
        return generate_expression(_rdiv,self,other)

    def __rtruediv__(self,other):
        """ Binary division

        (Called in response to 'other / self' with __future__.division.)
        """
        return generate_expression(_rdiv,self,other)

    def __rpow__(self,other):
        """ Binary power

        (Called in response to 'other ** self'.)
        """
        return generate_expression(_rpow,self,other)

    def __iadd__(self,other):
        """Binary addition

        (Called in response to 'self += other'.)
        """
        return generate_expression(_iadd,self,other)

    def __isub__(self,other):
        """ Binary subtraction

        (Called in response to 'self -= other'.)
        """
        return generate_expression(_isub,self,other)

    def __imul__(self,other):
        """ Binary multiplication

        (Called in response to 'self *= other'.)
        """
        return generate_expression(_imul,self,other)

    def __idiv__(self,other):
        """ Binary division

        (Called in response to 'self /= other'.)
        """
        return generate_expression(_idiv,self,other)

    def __itruediv__(self,other):
        """ Binary division

        (Called in response to 'self /= other' with __future__.division.)
        """
        return generate_expression(_idiv,self,other)

    def __ipow__(self,other):
        """ Binary power

        (Called in response to 'self **= other'.)
        """
        return generate_expression(_ipow,self,other)

    def __neg__(self):
        """ Negation

        (Called in response to '- self'.)
        """
        return generate_expression(_neg,self, None)

    def __pos__(self):
        """ Positive expression

        (Called in response to '+ self'.)
        """
        return self

    def __abs__(self):
        """ Absolute value

        (Called in response to 'abs(self)'.)
        """
        return generate_expression(_abs,self, None)


class NumericConstant(NumericValue):
    """An object that contains a constant numeric value.

    Constructor Arguments:
        value           The initial value.
    """

    __slots__ = ('value',) 

    def __init__(self, value):
        self.value = value
 
    def __getstate__(self):
        state = super(NumericConstant, self).__getstate__()
        for i in NumericConstant.__slots__:
            state[i] = getattr(self,i)
        return state
 
    def is_constant(self):
        return True

    def is_fixed(self):
        return True

    def __str__(self):
        return str(self.value)

    def to_string(self, ostream=None, verbose=None, precedence=0):
        if ostream is None:
            ostream = sys.stdout
        ostream.write(self.__str__())

    def __nonzero__(self):
        """Return True if the value is defined and non-zero."""
        if self.value:
            return True
        if self.value is None:
            raise ValueError("Numeric Constant: value is undefined")
        return False

    __bool__ = __nonzero__

    def __call__(self, exception=True):
        """Return the value of this object."""
        return self.value

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:         #pragma:nocover
            ostream = sys.stdout
        ostream.write(str(self))

# We use as_numeric() so that the constant is also in the cache
ZeroConstant = as_numeric(0)

# Why is this here???
#
# JDS: numvalue.py and expr.py have circular imports: the imports below
# from the expression system, and as_numeric and native_numeric_types
# from here.  Placing these imports here (after the body of the module)
# breaks the circular dependency.  
#
# FIXME: The correct solution is to refactor these two files into three
# files without a circular dependency.
from coopr.pyomo.base.expr import \
    generate_expression, generate_relational_expression, \
    _add, _sub, _mul, _div, _pow, _neg, _abs, _inplace, \
    _radd, _rsub, _rmul, _rdiv, _rpow, _iadd, _isub, _imul, _idiv, _ipow, \
    _lt, _le, _eq



import os

import pyutilib.th as unittest
from nose.tools import nottest
from coopr.pyomo.check import *
from coopr.pyomo.plugins.check.checker import *

currdir = os.path.dirname(os.path.abspath(__file__))

class MockChecker(object):
    checkCount = 0
    def check(self, runner, script, info):
        self.checkCount += 1
    def _checkerPackage(self):
        return 'mock'
    def resetCount(self):
        self.checkCount = 0

# Multiple inheritance, because otherwise something weird happens with the plugin system
# MockChecker MUST BE specified first so that its overridden methods take precedence
# See http://www.python.org/download/releases/2.3/mro/

class MockImmediateDataChecker(MockChecker, ImmediateDataChecker): pass
class MockImmediateTreeChecker(MockChecker, ImmediateTreeChecker): pass
class MockIterativeDataChecker(MockChecker, IterativeDataChecker): pass
class MockIterativeTreeChecker(MockChecker, IterativeTreeChecker): pass

class RunnerTest(unittest.TestCase):
    """
    Test the ModelCheckRunner class.
    """

    testScripts = [
        "print('Hello, world!')\n",
        "import sys\nsys.stdout.write('Hello, world!\\n')\n"
        "for i in range(10):\n\tprint(i)\n"
    ]

    def test_init(self):
        "Check that a ModelCheckRunner instantiates properly"

        runner = ModelCheckRunner()

        self.assertEqual([], runner.scripts)
        self.assertTrue(len(runner._checkers(all=True)) > 0)
        for c in runner._checkers:
            self.assertTrue(IModelChecker in c._implements.keys())

    def test_addScript(self):
        "Check that a runner handles its script list properly"

        runner = ModelCheckRunner()
        expectedScriptCount = 0

        for text in self.testScripts:
            self.assertEquals(expectedScriptCount, len(runner.scripts))

            script = ModelScript(text = text)
            runner.addScript(script)
            expectedScriptCount += 1

            self.assertEquals(expectedScriptCount, len(runner.scripts))
            self.assertTrue(script in runner.scripts)

    def test_run_immediate(self):
        "Check that a runner calls check() on an immediate checker"

        for text in self.testScripts:
            
            runner = ModelCheckRunner()
            script = ModelScript(text = text)
            runner.addScript(script)

            runner.run(checkers = {'mock':['MockImmediateDataChecker', 'MockImmediateTreeChecker']})

            for klass in [MockImmediateDataChecker, MockImmediateTreeChecker]:
                mockChecker = list(filter((lambda c : c.__class__ == klass), runner._checkers()))[0]
                self.assertEqual(1, mockChecker.checkCount)
                mockChecker.resetCount()

    def test_run_iterative(self):
        "Check that a runner calls check() on an iterative checker"

        for text in self.testScripts:
            
            runner = ModelCheckRunner()
            script = ModelScript(text = text)
            runner.addScript(script)

            runner.run(checkers = {'mock':['MockIterativeDataChecker', 'MockIterativeTreeChecker']})

            for klass in [MockIterativeDataChecker, MockIterativeTreeChecker]:
                mockChecker = list(filter((lambda c : c.__class__ == klass), runner._checkers()))[0]
                self.assertTrue(mockChecker.checkCount >= 1)
                mockChecker.resetCount()


if __name__ == "__main__":
    unittest.main()

